@@ -30,6 +30,7 @@ from scipy.interpolate import make_interp_spline
 
 from ml_processor.configuration import config
 from ml_processor.jsonSerializer import NpEncoder
+from ml_processor.data_prep import balance_data
 
 sns.set_style('whitegrid')
 warnings.filterwarnings('ignore')
@@ -74,12 +75,19 @@ class model_training:
 
     search_space : dic (default = None)
         Range of values for the different parameters to tune.
+
+    balance_data : string (default=None, options=[None, "train", "both"])
+        Whether to balance data according to the distribution of the labels or not.
+            • None - Doesn't balance data.
+            • train_only - Balances only training data
+            • both - Balances both training and testing data
     
     """
 
     
     def __init__(self, df, features, target="target", params_prop=0.25, test_size=0.33, hyperparams=None, 
-           method="hyperopt", classifier="xgboost", eval_metric="recall", search_space=None
+           method="hyperopt", classifier="xgboost", eval_metric="recall", search_space=None,
+           balance_data = None
            ):
         
         self.data = df
@@ -97,6 +105,8 @@ class model_training:
         self.eval_metric = eval_metric
         
         self.search_space = search_space
+
+        self.balance_data = balance_data
         
         self.cwd = os.getcwd()
         
@@ -214,7 +224,7 @@ class model_training:
         return self.model
 
 
-    def model_evaluation(self, model = None, features = None, target = "target", train_set = pd.DataFrame(), 
+    def model_evaluation(self, model = None, features = None, target = None, train_set = pd.DataFrame(), 
         validation_set = pd.DataFrame(), test_set = pd.DataFrame(), main_set = "test", loc_path = None
         ):
 
@@ -231,7 +241,7 @@ class model_training:
         features : list or array-like (default=None)
             Features used for fitting the model
 
-        target : string (default="target")
+        target : string (default=None)
             COlumn with labels
 
         train_set : pandas.Dataframe (default=pandas.DataFrame())
@@ -260,6 +270,9 @@ class model_training:
         
         if  not features:
             features = self.features
+
+        if not target:
+            target = self.target
         
         if train_set.empty:
             train_set = pd.concat([self.X_train, self.y_train], axis=1)
@@ -284,7 +297,7 @@ class model_training:
         return model_perf.make_plots()
 
 
-    def shap_summary_plot(self, model=None, eval_set=pd.DataFrame(), save_plot=True, save_path=None, plot_size=(15,20) ):
+    def shap_summary_plot(self, model=None, eval_set=pd.DataFrame(), max_display=None, save_plot=True, save_path=None, plot_size=(15,20) ):
 
         """
 
@@ -296,6 +309,9 @@ class model_training:
         eval_set : pandas.DataFrame (default=pandas.DataFrame())
             Data set to use for model evaluation
 
+        max_display : int (default=None)
+            Number of columns to plot
+
         save_plot : boolean (default=True)  
             Whether to save plot or not
 
@@ -313,11 +329,14 @@ class model_training:
         if eval_set.empty:
             eval_set = self.X_train
 
+        if not max_display:
+            max_display = len(eval_set.columns)
+
         if save_plot:
             if not save_path:
                 save_path = os.path.join(f"{self.classifier}_model_artefacts", "shap_summary_plot.png")
 
-        model_interp_shap(model=model, eval_set=eval_set, save_path=save_path, plot_size=plot_size )
+        model_interp_shap(model=model, eval_set=eval_set, max_display=max_display, save_path=save_path, plot_size=plot_size )
 
 
     def _parameter_tunning(self):
@@ -468,7 +487,32 @@ class model_training:
         
         """
 
-        self.X_train, self.X_test, self.y_train, self.y_test = split_data(data=self.data, target=self.target, test_size=self.test_size)
+        if not self.balance_data:
+
+            self.X_train, self.X_test, self.y_train, self.y_test = split_data(data=self.data, target=self.target, test_size=self.test_size)
+        
+        elif self.balance_data == "both":
+            
+            data_balanced = balance_data(target=self.target, data=self.data)
+
+            self.log.info('Data balanced acording to minority size of the labels')
+            
+            self.X_train, self.X_test, self.y_train, self.y_test = split_data(data=data_balanced, target=self.target, test_size=self.test_size)
+        
+        elif self.balance_data == "train_only":
+
+            self.X_train, self.X_test, self.y_train, self.y_test = split_data(data=self.data, target=self.target, test_size=self.test_size)
+
+            train_set = pd.concat([self.X_train, self.y_train], axis=1)
+
+            train_set_balanced = balance_data(target=self.target, data=train_set)
+
+            self.log.info('Training set balanced acording to minority size of the labels')
+
+            self.X_train = train_set_balanced[self.features]
+
+            self.y_train =  train_set_balanced[self.target]
+
         
         self.log.info('Splitting data into training and testing sets completed')
         
@@ -686,7 +730,7 @@ class plot_model_perf:
             width = 12
             rows = 2
         
-        fig = plt.figure(figsize=(25, width), constrained_layout=True)
+        fig = plt.figure(figsize=(40, width), constrained_layout=True)
 
         spec2 = gridspec.GridSpec(ncols=5, nrows=rows, figure=fig)
         
@@ -775,7 +819,7 @@ def split_data(data, target="target", test_size=0.33):
 
 
 
-def model_interp_shap(model, eval_set, save_path=None, plot_size=(10,10)):
+def model_interp_shap(model, eval_set, max_display=None, save_path=None, plot_size=(10,10)):
 
     """
 
@@ -787,6 +831,9 @@ def model_interp_shap(model, eval_set, save_path=None, plot_size=(10,10)):
     eval_set : pandas.DataFrame (default=pandas.DataFrame())
         Data set to use for model evaluation
 
+    max_display : int (default=None)
+        Number of columns to plot
+
     save_path : string (default = None)
         Path to save plot to.
 
@@ -796,7 +843,8 @@ def model_interp_shap(model, eval_set, save_path=None, plot_size=(10,10)):
     """
 
     features = eval_set.columns
-    num_cols = len(features)
+    if not max_display:
+        max_display = len(features)
     
     if isinstance(model, xgb.sklearn.XGBClassifier):
         
@@ -804,7 +852,7 @@ def model_interp_shap(model, eval_set, save_path=None, plot_size=(10,10)):
         
         shap_values = explainer.shap_values(eval_set)
 
-        shap.summary_plot(shap_values, eval_set, max_display=num_cols, show=False, plot_size=plot_size)
+        shap.summary_plot(shap_values, eval_set, max_display=max_display, show=False, plot_size=plot_size)
 
         if save_path:
             plt.savefig(save_path, bbox_inches='tight',dpi=100)
@@ -823,7 +871,7 @@ def model_interp_shap(model, eval_set, save_path=None, plot_size=(10,10)):
         
         shap_values.base_values = shap_values.base_values[:,1]
         
-        shap_plot = shap.plots.beeswarm(shap_values, max_display=num_cols, show=False, plot_size=plot_size)
+        shap_plot = shap.plots.beeswarm(shap_values, max_display=max_display, show=False, plot_size=plot_size)
 
         if save_path:
             plt.savefig(save_path, bbox_inches='tight',dpi=100)
@@ -1150,12 +1198,11 @@ def search_space_generator(classifier="xgboost"):
     space = {
         "xgboost" : {
             'n_estimators': hp.choice('n_estimators', range(100,1500,100)),
-            'learning_rate': hp.choice('learning_rate', np.arange(0.01, 0.5, 0.01)),
             'max_depth': hp.choice('max_depth', range(2, 15, 1)),
+            'learning_rate': hp.choice('learning_rate', [0.0001, 0.001, 0.01, 0.1, 1.0]),
             'subsample': hp.choice('subsample', np.arange(0.1, 1.1, 0.1)),
             'colsample_bytree': hp.choice('colsample_bytree', np.arange(0.1, 1.1, 0.1)),
-            'gamma': hp.choice('gamma', np.arange(0.1, 0.6, 0.1)),
-            'reg_alpha': hp.choice('reg_alpha', [1e-5, 1e-2, 0.1, 1, 10, 100]),
+            'gamma': hp.choice('gamma', np.arange(0.0, 1.0, 0.25)),
             'reg_lambda': hp.choice('reg_lambda', [1e-5, 1e-2, 0.1, 1, 10, 100]),
             'objective': 'binary:logistic',
             'booster': 'gbtree',
@@ -1260,7 +1307,7 @@ def _plot_prediction(y_test, pred_class, pred_prob, ax, dataset="test"):
     
     df_pred = pd.DataFrame({'actual_class': y_test, 'pred_class': pred_class, 'pred_prob': pred_prob})
     
-    sns.kdeplot(data=df_pred, x='pred_prob', hue='pred_class', fill=True, legend=False, palette=palette, ax=ax)
+    sns.kdeplot(data=df_pred, x='pred_prob', hue='actual_class', fill=True, legend=False, palette=palette, ax=ax)
     
     ax.set_facecolor('#274472')
     
